   module flamsteed_subs

       use opt_extr
       use ark_file_io
       use define_star

       implicit none

       contains

       subroutine doubles(seeing, star, nstar)

         ! Detects close pairs (i.e. stars within nsee*seeing) and
         ! removes one of the two.

         real, intent(in) :: seeing
         type(a_star), dimension(:), intent(inout) :: star
         integer, intent(inout) :: nstar

         type(a_star), allocatable, dimension(:) :: star2
         integer :: istar2, i, j
         real :: work

         real, parameter :: nsee=1

         ! Remove doubles.
         allocate(star2(nstar))
         istar2=1
         star2(1)=star(1)
         compare: do i=2, nstar
           do j=1, istar2
             if (abs(star(i)%x-star2(j)%x) < real(nsee)*seeing) then
               if (abs(star(i)%y-star2(j)%y) < real(nsee)*seeing) then
                 work=(star(i)%y-star2(j)%y)**2.0 + (star(i)%x-star2(j)%x)**2.0
                 if (work < (real(nsee)*seeing)**2.0) then
                   ! Remove one star.
                   cycle compare  
                 end if
               end if
             end if
           end do
           istar2=istar2+1
           star2(istar2)=star(i)
         end do compare
         nstar=istar2
         star(1:nstar)=star2(1:nstar)
         deallocate(star2)

       end subroutine doubles

       subroutine back(data, low, high, ibox, noise_map, median_noise, &
         pix_flg)

         ! This subroutine background subtracts the data, and returns
         ! a noise map.

         ! The data to be background subtracted.
         real, dimension(:,:), intent(inout) :: data
         ! The active area of the CCD.
         integer, dimension(2), intent(in) :: low, high
         ! The size of the box to be used.
         integer, intent(in) :: ibox
         ! And the noise map
         real, dimension(:,:), allocatable, intent(out) :: noise_map
         ! Median sky noise level.
         real :: median_noise
         ! The pixel flags.
         character, dimension(:,:), intent(in) :: pix_flg
         
         real :: skyerr, median_sky
         integer :: i, j, iaxis1, iaxis2, ncount
         ! The values of the sky and sky noise at selected points.
         real, dimension(:,:), allocatable :: skycnt, skynos
         logical, dimension(:,:), allocatable :: bad
         ! Conversion of the array element numbers into image pixel numbers.
         integer, dimension(:,:), allocatable :: axis
         ! The number of boxes in each direction
         integer, dimension(2) :: nboxes
         character :: cflag
         real :: work
         ! For finding median sky and noise.
         real, allocatable, dimension(:) :: noise_buf, sky_buf

         ! Interpolation numbers.
         real :: t, u

         nboxes=(high-low+1)/ibox
         allocate(axis(maxval(nboxes),2))
         allocate(skycnt(nboxes(1),nboxes(2)))
         allocate(skynos(nboxes(1),nboxes(2)))
         allocate(bad(nboxes(1),nboxes(2)))
         allocate(noise_buf(nboxes(1)*nboxes(2)), sky_buf(nboxes(1)*nboxes(2)))
         bad=.false.

         call setbug()
         ncount=0
         do i=1, nboxes(1)
           iaxis1 = (low(1)-1) + (i-1)*ibox + ibox/2
           axis(i,1)=iaxis1
           do j=1, nboxes(2)
             iaxis2 = (low(2)-1) + (j-1)*ibox + ibox/2
             axis(j,2)=iaxis2
             !print*, '@ Fitting sky centered on ', real(iaxis1), real(iaxis2) 
             call skyfit(data, real(iaxis1), real(iaxis2), 1.0, ibox, &
             low, high, pix_flg, skycnt(i,j), skyerr, skynos(i,j), cflag)
             write(23,*) iaxis1, iaxis2, skynos(i,j), skycnt(i,j), cflag
             if (cflag == 'B') then
               ! Usually means the area is seriously stuffed by a bright
               ! star.  Setting the noise huge means there
               ! won't be any significant events in the region, or anything
               ! that uses it for interpolation.
               bad(i,j)=.true.
               !print*, 'Sky around ', iaxis1, iaxis2, ' is bad.'
             else
               ncount=ncount+1
               noise_buf(ncount)=skynos(i,j)
               sky_buf  (ncount)=skycnt(i,j)
             end if
           end do
         end do

         print*, count(bad), ' out of ', count(bad)+ncount, &
         ' sky regions were bad.'

         median_noise=median(noise_buf, ncount)
         median_sky  =median(sky_buf,   ncount)
         
         where(bad) 
           skycnt=median_sky
           skynos=median_noise
         end where

         deallocate(sky_buf, noise_buf)

         print*, 'Creating noise map, and sky subtracting.'
         allocate(noise_map(size(data,1),size(data,2)))
         noise_map=0.0
         do iaxis1=low(1), high(1)
           i=(iaxis1-(low(1)-1)+ibox/2)/ibox
           if (i < 1) then
             t=0.0
             i=1
           else if (i > nboxes(1)-1) then
             t=1.0
             i=nboxes(1)-1
           else
             t = real(iaxis1-axis(i,1))/real(axis(i+1,1)-axis(i,1))
           end if
           do iaxis2=low(2), high(2)
             j=(iaxis2-(low(2)-1)+ibox/2)/ibox
             if (j < 1) then
               u=0.0
               j=1
             else if (j > nboxes(2)-1) then
               u=1.0
               j=nboxes(2)-1
             else
               u = real(iaxis2-axis(j,2))/real(axis(j+1,2)-axis(j,2))
             end if
             work =                            &
             (1.0-t)*(1.0-u)*skycnt(i,  j)   + &
                  t *(1.0-u)*skycnt(i+1,j)   + &
                  t *     u *skycnt(i+1,j+1) + &
             (1.0-t)*     u *skycnt(i,  j+1)
             data(iaxis1,iaxis2) = data(iaxis1,iaxis2) - work             
             ! Calculate the noise map.
             noise_map(iaxis1,iaxis2) =        &
             (1.0-t)*(1.0-u)*skynos(i,  j)   + &
                  t *(1.0-u)*skynos(i+1,j)   + &
                  t *     u *skynos(i+1,j+1) + &
             (1.0-t)*     u *skynos(i,  j+1)
           end do
         end do
         
         ! Now, set anything with a bad sky to zero counts.
         do i=1, nboxes(1)
           do j=1, nboxes(2)
             if (bad(i,j)) &
             data(max(1,axis(i,1)-ibox/2):min(axis(i,1)+ibox/2,high(1)), &
                  max(1,axis(j,2)-ibox/2):min(axis(j,2)+ibox/2,high(2))) &
             =0.0
           end do
         end do
         
         deallocate(skycnt, skynos) 

       end subroutine back

       subroutine sig(data, low, high, sigma, seeing, &
         noise_map, pix_flg, bin_sig_map)

         ! This subroutine returns a significance map.

         ! The data to be background subtracted.
         real, dimension(:,:), intent(inout) :: data
         ! The active area of the CCD.
         integer, dimension(2), intent(in) :: low, high
         ! The size of the box to be used.
         ! The sigma events to search for.
         real, intent(in) :: sigma
         ! The fwhm seeing in pixels.
         real, intent(in) :: seeing
         ! And the noise map
         real, dimension(:,:), intent(in) :: noise_map
         ! The pixel flags.
         character, dimension(:,:), intent(in) :: pix_flg
         
         ! The significance map.
         logical, dimension(:,:), intent(out) :: bin_sig_map

         ! The box size used in smoothing
         integer :: sbox

         integer :: iaxis1, iaxis2, jaxis1, jaxis2
         real :: bin_sig_pix
         ! The arrays of star positions and peak flux.
         integer :: icount

         sbox = (seeing-1)/2
         print*, 'Creating significance map using box size of', 2*sbox + 1
         bin_sig_map = .false.
         do iaxis1=low(1)+sbox, high(1)-sbox
           do iaxis2=low(2)+sbox, high(2)-sbox
             ! Filter the image by adding together a block of pixels, but
             ! ignore those which are flagged.
             icount=0
             bin_sig_pix = 0.0
             do jaxis1=iaxis1-sbox, iaxis1+sbox
               do jaxis2=iaxis2-sbox, iaxis2+sbox
                 if (jaxis1==iaxis1 .and. jaxis2==iaxis2) then
                   bin_sig_pix=bin_sig_pix+data(jaxis1,jaxis2)
                   icount=icount+1
                 else
                   if (pix_flg(jaxis1,jaxis2) == 'O') then
                     icount=icount+1
                     bin_sig_pix = bin_sig_pix + data(jaxis1,jaxis2)
                   end if
                 end if
               end do
             end do
             bin_sig_pix=bin_sig_pix/real(icount)
             bin_sig_pix = real(2*sbox+1)*bin_sig_pix / noise_map(iaxis1, iaxis2)
             if (bin_sig_pix > sigma) bin_sig_map(iaxis1,iaxis2) = .true.
           end do
         end do
         
         !where (bin_sig_map)
         !  data=1.0
         !elsewhere
         !  data=0.0
         !end where
         !return

       end subroutine sig

       subroutine filter(x_sig, y_sig, n_sig, &
       high, data, pix_flg, star, nstar)

         ! Takes a list of significant events, and then uses a connectivity
         ! condition to make a list of stars.  If there are already stars
         ! in the array, then it tags new stars on the end, and ignores any
         ! group of events which belongs to an already extant star.  This 
         ! means that on its first calling, nstar MUST be set to zero.

         ! The list of pixel positions above n sigma.
         integer, dimension(:), intent(in) :: x_sig, y_sig
         ! How many such pixels are there?
         integer, intent(in) :: n_sig
         ! The highest good pixels in each dimension.
         integer, dimension(2), intent(in) :: high
         ! The data array.
         real, dimension(:,:), intent(in) :: data
         ! The pixel flags.
         character, dimension(:,:), intent(in) :: pix_flg
         
         ! The current list of identified stars.
         type(a_star), dimension(:), intent(inout) :: star
         ! And the number of them.
         integer, intent(inout) :: nstar

         ! This points to the next pixel in the list.
         integer, allocatable, dimension(:) :: pointer
         ! Two pixel counters.
         integer :: ipixel, jpixel

         ! The heads of each list.
         integer, allocatable, dimension(:) :: start

         ! The picture of the image.  This will be zero in sky regions, and
         ! contain the number of the object in each pixel of that object.
         integer, dimension(-1:high(1),-1:high(2)) :: piccy
         ! real, dimension(:,:), allocatable :: axdata

         ! The last list of pixel numbers to be started, which ends up
         ! being the total number of lists.  And a list counter.
         integer :: nlist, list
         ! The image numbers of the pixel to the left and below the current
         ! pixel in piccy.
         integer :: down, left
         ! A pixel counter.
         integer :: npixel
         ! For findng the mean positions of stars.
         real :: xsum, ysum
         ! For finding the peak positions of the stars.
         real :: peak_xpos, peak_ypos, value, peak
         integer :: ixpos, iypos
         ! Keepers for x_sig and y_sig.
         integer :: ix_sig, iy_sig
         integer :: istar
         ! The number of stars on entry.
         integer :: nstar_in
         character :: flg

         ! Allocate the array of pointers,
         allocate(pointer(n_sig))
         ! and set them all to mean end-of-list.
         pointer=0

         ! Allocate the starts of each list.
         allocate(start(n_sig))
         
         ! To begin with there are no images in the picture.
         piccy=0
         
         ! And save the number of stars on entry.
         nstar_in = nstar
         

         nlist=0
         do ipixel=1, n_sig
           left=0
           down=0
           ix_sig=x_sig(ipixel)
           iy_sig=y_sig(ipixel)
           left=piccy(ix_sig-1,iy_sig)
           down=piccy(ix_sig,iy_sig-1)
           if (left==0 .and. down==0) then
             ! A whole new image.
             nlist=nlist+1
             ! Make a starter point here.
             start(nlist)=ipixel
             ! and fill in the piccy.
             piccy(ix_sig,iy_sig)=nlist
           else if (left==0 .and. down/=0) then
             piccy(ix_sig,iy_sig)=down
             ! Make the pointer point to the star which was previously head of
             ! the list.
             pointer(ipixel)=start(down)
             ! And make the start point to this pixel.
             start(down)=ipixel
           else if (left/=0 .and. down==0) then
             piccy(ix_sig,iy_sig)=left
             pointer(ipixel)=start(left)
             start(left)=ipixel
           else if (left == down) then
             ! An easy case.
             piccy(ix_sig,iy_sig)=down
             pointer(ipixel)=start(down)
             start(down)=ipixel
           else
             ! This joins two images.  Go down the left image until you reach
             ! the head of the list.
             jpixel=start(left)
             going_down: do 
               piccy(x_sig(jpixel), y_sig(jpixel))=down
               if (pointer(jpixel) == 0) exit going_down
               jpixel=pointer(jpixel)
             end do going_down
             ! Connect the end of left to the start of down. 
             pointer(jpixel)=start(down)
             ! Now put the new pixel at the start of the list.
             pointer(ipixel)=start(left)
             ! Point to it from the head of the list.
             start(down)=ipixel
             ! And place a number in the piccy.
             piccy(ix_sig,iy_sig)=down
             ! Nullify left.
             start(left)=0
           end if
         end do

         print*, 'Consolidating pixels into stars.'

         consolidate: do list=1, nlist
       
           ! Ignore lists with zero pixels.
           if (start(list) == 0) cycle consolidate

           xsum=0.0
           ysum=0.0
           ixpos=0.0
           iypos=0.0
           npixel=0
           ! Since the data are sky subtracted silly things can happen.
           peak=-huge(peak)
           jpixel=start(list)
           flg='O'
           walk: do
             ! First, is this pixel already marked as a star?
             if (nstar > 0) then
               do istar=1, nstar_in
                 if (nint(star(istar)%x) == x_sig(jpixel)) then
                   if (nint(star(istar)%y) == y_sig(jpixel)) then
                     ! Yes it is an already known star.
                     cycle consolidate
                   end if
                 end if
               end do
             end if               
             value=data(x_sig(jpixel), y_sig(jpixel))
             if (pix_flg(x_sig(jpixel), y_sig(jpixel)) == 'O') then
               xsum=xsum+real(x_sig(jpixel))*value
               ysum=ysum+real(y_sig(jpixel))*value
             else
               flg=pix_flg(x_sig(jpixel), y_sig(jpixel))
             end if
             ixpos=ixpos+x_sig(jpixel)
             iypos=iypos+y_sig(jpixel)
             npixel=npixel+1
             if (value > peak) then
               peak=value
               peak_xpos=x_sig(jpixel)
               peak_ypos=y_sig(jpixel)
             end if
             if (pointer(jpixel) == 0) exit walk
             jpixel=pointer(jpixel)
           end do walk
           if (nstar+1 > size(star)) then
             print*, 'Number of stars is greater than available array'
             print*, 'size in filter.  Array size is ', size(star)
             print*, 'Truncating star list.'
             exit consolidate
           end if
           nstar=nstar+1
           ! Find the position.  This is more robust than the first moment.
           if (flg=='O' .or. npixel==0) then
             ! Take the brightest pixel to be the position.
             star(nstar)%x=peak_xpos
             star(nstar)%y=peak_ypos
           else
             ! Take the mean position of all pixels.
             star(nstar)%x=real(ixpos)/real(npixel)
             star(nstar)%y=real(iypos)/real(npixel)
           end if
           ! Give the star a number.
           star(nstar)%id=nstar
           ! And a flag.
           star(nstar)%col(1)%flg='O'//flg
         end do consolidate
       
         deallocate(pointer,start)
           
       end subroutine filter

       real function median(srtbuf, nfile)

         integer, intent(in) :: nfile
         real, dimension(nfile) :: srtbuf

         integer :: k, l, m
         real :: aval
                  
         loop1: do k=2, nfile
           aval=srtbuf(k)
           do l=1, k-1
             if (aval > srtbuf(l)) then
               do m=k, l+1, -1
                 srtbuf(m)=srtbuf(m-1)
               end do
               srtbuf(l)=aval
               cycle loop1
             endif
           end do
         end do loop1

         median=srtbuf(nfile/2 + 1)
         if (mod(nfile,2) .eq. 0) median=(median+srtbuf(nfile/2))/2.0

       end function median
      
   end module flamsteed_subs
