     program opphot                         

        ! Carries out optimal extraction of photometry using the method
        ! described by Naylor (1998).  You need to run the code twice.
        ! The first pass will produce accurate positions of the stars
        ! (in stars.pos_new), which should then be fed into the second
        ! run.

        ! By Tim Naylor.

        ! Version 1.1
        !   Now fails elegantly when psf.pos or stars.pos is incorrectly
        !   formatted.
        ! Version 2.0
        !   The position angle of the Gaussian PSF allowed to run free.
        ! Version 2.1
        !   1) Small bug cured which referred to a non-existent element of
        !   dpsf when writing a comment to an output file, and to a 
        !   non-existant element of dpos when writing to the screen.
        !   2) comp reset to .false. for each new star.
        ! Version 2.2
        !   Changed format of .log file.
        ! Version 2.3
        !   More comments to aid porting, and default values of aperture and
        !   mask radius set using fwhm from first image.
        ! Version 2.4
        !   Added errors in position into .log file.
        ! Version 2.5
        !   Increased to 99,999 output file names for Monica
        ! Version 2.6
        !   Log file changed to imagename.opt file, and imagename.dat added.
        !   These are files used for colour-magnitude diagrams, and hence
        !   if there are more than 100 stars, the star????? files are not
        !   written.
        ! Version 2.7
        !   Minor changes so that it compiles with the NAG Fortran 95
        !   compiler, as well as with the DEC Fortran 90 compiler.
        ! Version 2.?
        !   1) Allows user to skip aperture photometry.
        !   2) Adaptively sets the clipping radius to be a mutliple of the 
        !      FWHM of the PSF star in each frame.
        !   3) Allows the user to skip the optimal photometry.
        !   4) Checks for a high skewness in the sky histogram.
        !   5) Signals problem measurements only through iflag, never by
        !      setting the error negative.

        ! This module contains all the ARK I/O routines.
        use ark_file_io
        ! This module contains the optimal extraction stuff.
        use opt_extr
        ! This is a stand alone routine.
        use detector

        implicit none

        ! For the list of PSF stars.
        character(len=50) :: psfpos
        integer :: npsf, ipsf
        real, allocatable, dimension(:) :: xpsf, ypsf, dpsf
        integer, allocatable, dimension(:) :: nopsf
                          
        ! For the list of target stars.
        character(len=50) :: strpos
        integer :: nstar                   
        real, allocatable, dimension(:) :: xcoor, ycoor, dpos
        integer, allocatable, dimension(:) :: nostar
        ! The output file for these stars.
	character(len=50), allocatable, dimension(:) :: ofopt, ofap
        real, allocatable, dimension(:) :: xpos_new, ypos_new, tot_sig_nos
        integer, allocatable, dimension(:) :: npos_new

        ! The positions of stars shifted from the current reference frame
        ! into the new one.
        real, allocatable, dimension(:) :: xpos0, ypos0

        integer :: i, j, istar, jstar
        ! An answer string
        character(len=50) :: answer           
        logical :: there
        integer :: iunit1, iostat, iunit2, iunit4, iunit5
        real :: flux, error
        real :: best_n, peak

        ! For the image.
        integer :: naxis(2)
        real, allocatable, dimension(:,:) :: data, axdata
        ! The active area of the detector.
        integer :: low(2), high(2), inhigh(2)

        ! Is there a companion, and what is its position?
        logical :: comp
        real :: xcomp, ycomp
                                        
        ! For skyfit.
        real :: skycnt, skyerr, skynos
        integer :: iflag, ibox
        logical :: sky_stuffed

        ! For the fitting routine.
        ! The first 3 parameters, which define the shape of the function.
        real, dimension(3) :: shape_par

        ! The gain.
        real :: adu
        ! The number of the star whose signal-to-noise is to be optimised.
        integer :: iopt
        ! And its peak counts.
        real :: optnrm

        ! For offsets.dat
        character(len=50) :: offset, ifname
        real :: xoff, yoff, sky, theoff, fwhm
        integer ntoe
                              
        ! For times.dat
        character(len=50) :: timfil
        real, allocatable, dimension(:) :: time, dtime
        character(len=50), allocatable, dimension(:) :: timnam
        integer ntim
        ! The times for the current frame.
        real :: timec, dtimec

        ! For dealing with bad pixels.
        real, allocatable, dimension(:) :: xbpl, ybpl
        integer :: nbpl
        real :: badpix

        ! The shifted position of psf and target star
        real :: xpos, ypos, r1, theta1
        ! And the fitted position.
        real :: xfit, yfit
        ! And its error.
        real :: xerr, yerr

        ! For clipping the edge of the weight region.
        real :: cliprad
        ! How many times the FWHM cliprad is set.
        real :: clip_fwhm

        ! The aperture for aperture photometry.
        real :: aprad
                                   
        ! The log files for the optimal and aperture photometry.
        character(len=50) :: oflog
        integer :: logopt_unit, logap_unit

        ! Unused arguments to rd_detector
        real :: satur, read
                            
        ! Variables for debugging                     
	logical :: debug
        common /bug/ debug

        ! ***************************************************************
        ! Actual program starts here.
        ! ***************************************************************
                     
        call setbug()

        if (debug) call opt_debug_on()


        ! ***************************************************************
        ! Read the co-ordinates for each psf star.
        ! ***************************************************************

        print*, ' '
        print*, '* First this program requires a list of possible PSF'
        print*, '* stars.  This should be a file where each line has'
        print*, '* a star number, followed by its X and Y co-ordinates'
        print*, '* in the reference data frame, and the radius to be'
        print*, '* searched for the center of the star.  If this last'
        print*, '* number is zero the star position is taken as fixed.'
        print*, '* (The first 3 lines of the file are ignored.)'
        print*, '* (Type end if you don''t want optimal photometry.)'

        psffil: do 
          print*, '> Give the name of this file (<cr>=psf.pos).'
          psfpos=' '
          read (*, '(a)') psfpos
          if (psfpos=='end' .or. psfpos=='END') then
            clip_fwhm = -1.0
            exit psffil
          else
            clip_fwhm=1.0
          end if
          if (psfpos .eq. ' ') psfpos='psf.pos'
          inquire(file=psfpos, exist=there)
          if (.not. there) then
            print*, '* File not found.'
          else
           exit psffil
          end if
        end do psffil

        if (clip_fwhm > 0.0) then
          ! Open the file with the position(s) of the psf star(s) using a 
          ! machine independent ARK routine.  It returns the number of the
          ! unit the file is opened on.
          call arkopn(iunit1, psfpos, ' ', 'OLD', 'READONLY', &
          'FORMATTED', ' ', 0)
          read(iunit1,*,iostat=iostat)
          read(iunit1,*,iostat=iostat)
          read(iunit1,*,iostat=iostat)
          npsf=0
          ctpsf: do
            read(iunit1, '(a)', iostat=iostat) psfpos
            if (iostat < 0) exit ctpsf
            npsf=npsf+1
          end do ctpsf
          allocate(nopsf(npsf))
          allocate(xpsf(npsf))
          allocate(ypsf(npsf))
          allocate(dpsf(npsf))
          rewind (iunit1)
          read(iunit1,*,iostat=iostat)
          read(iunit1,*,iostat=iostat)
          read(iunit1,*,iostat=iostat)
          psfin: do i=1, npsf
            read(iunit1,*,iostat=iostat) nopsf(i), xpsf(i), ypsf(i), dpsf(i)
            if (iostat /= 0) then
              print*, 'Error in psf star position file at line ', i+3
              stop
            end if
          end do psfin
          close(iunit1)
                                                  
          print*, '* ', npsf, ' psf star positions read.'

        end if

        ! *******************************************************************
        ! Read the co-ordinates for each target star, and make a file for it.
        ! *******************************************************************

        print*, ' '
        print*, '* Next the program requires a list of the target'
        print*, '* stars.  This should be a file where each line has'
        print*, '* a star number, followed by its X and Y co-ordinates'
        print*, '* in the reference data frame, and the radius to be'
        print*, '* searched for the center of the star.  If this last'
        print*, '* number is zero the star position is taken as fixed.'
        print*, '* (The first 3 lines of the file are ignored.)'

        strfil: do
          print*, '> Give the name of this file (<cr>=stars.pos).'
          strpos=' '
          read (*, '(a)') strpos
          if (strpos == ' ') strpos='stars.pos'
          inquire(file=strpos, exist=there)
          if (.not. there) then
            print*, '* File not found.'
          else
           exit strfil
          end if
        end do strfil

        ! Open the file with the position(s) of the star(s) using a 
        ! machine independent ARK routine.  It returns the number of the
        ! unit the file is opened on.
        call arkopn(iunit1, strpos, ' ', 'OLD', 'READONLY', &
        'FORMATTED', ' ', 0)
	read(iunit1,*,iostat=iostat)
	read(iunit1,*,iostat=iostat)
	read(iunit1,*,iostat=iostat)
        nstar=0
        ctstr: do   
          read(iunit1, '(a)', iostat=iostat) strpos
          if (iostat < 0) exit ctstr
          nstar=nstar+1
        end do ctstr
        allocate(nostar(nstar))
        allocate(xcoor(nstar))
        allocate(ycoor(nstar))
        allocate(dpos(nstar))
        rewind (iunit1)
	read(iunit1,*,iostat=iostat)
	read(iunit1,*,iostat=iostat)
	read(iunit1,*,iostat=iostat)
	strin: do i=1, nstar
	  read(iunit1,*,iostat=iostat) nostar(i), xcoor(i), ycoor(i), dpos(i)
          if (iostat /= 0) then
            print*, 'Error in star position file at line ', i+3
            stop
          end if
        end do strin
        close(iunit1)

        print*, '* ', nstar, ' star positions read.'
        if (nstar >= 100) print*, '* Files will not be made for each star.'

        ! Allocate lots more things that need to know the number of stars.
        if (nstar < 100) then
          allocate(ofopt(nstar))
          allocate(ofap(nstar))
        end if
        allocate(xpos_new(nstar))
        allocate(ypos_new(nstar))
        allocate(npos_new(nstar))
        allocate(tot_sig_nos(nstar))
        allocate(xpos0(max(nstar,npsf)))
        allocate(ypos0(max(nstar,npsf)))

        ! ***************************************************************
        !                   Open offsets.dat
        ! ***************************************************************

        print*, ' '
        print*, '* This program also requires a file which has an'
        print*, '* image file name on one line, followed by the' 
        print*, '* offsets of that image on the next.'
        offsets: do
          print*, '> Give this file name (<cr>=offsets.dat).'
          offset=' '
          read (*, '(a)') offset
          if (offset .eq. ' ') offset='offsets.dat'
          inquire(file=offset, exist=there)
          if (.not. there) then
            print*, '* File not found.'
          else
            exit offsets
          end if
        end do offsets

        ! Open the file of offsets with between the data frames using a
        ! machine independent ARK routine.  It returns the number of the
        ! unit the file is opened on.
        call arkopn(iunit1, offset, ' ', 'OLD', 'READONLY', &
        'FORMATTED', ' ', 0)
        ! read the first couple of lines so we get a clue as to the
        ! seeing.
	read(iunit1,'(a)',iostat=iostat) ifname
        if (iostat < 0) then
          print*, 'Error reading offsets.dat file'
          stop
        end if
	read(iunit1,*,iostat=iostat) xoff, yoff, sky, theoff, ntoe, fwhm
        if (iostat < 0) then
          print*, 'Error reading offsets.dat file'
          stop
        end if
        ! But now rewind the file.
        rewind iunit1


        if (clip_fwhm > 0.0) then

        ! ***************************************************************
        !                   Open seeing.dat
        ! ***************************************************************

          ! Create the file which records the seeing in each frame using a
          ! machine independent ARK routine.  It returns the number of the
          ! unit the file is opened on.
          call arkopn(iunit5, 'seeing.dat', ' ', 'NEW', 'OVERWRITE', &
          'FORMATTED', ' ', 0)
          rewind (iunit5)
          write(iunit5,*) '! Column 1 is times from times.dat file.'
          write(iunit5,*) '! Column 2 is geometric mean of X and Y FWHM'
          write(iunit5,*) '! of PSF star, in pixels.'

        end if

        ! ***************************************************************
        !        Read the file of observation and exposure times.
        ! ***************************************************************

        print*, ' '
        print*, '* Next we require a file which has the image file'
        print*, '* name on one line, and the observation time and'
        print*, '* and half the exposure length on the next (in any'
        print*, '* units you like, they are just copied to the output).'
        times: do
          print*, &
          '> Give the name of this file ("end"=no file, <cr>=times.dat).'
          timfil=' '
          read (*, '(a)', iostat=iostat) timfil
          if (iostat < 0) exit times
          if (timfil=='END' .or. timfil=='end') exit times
          if (timfil == ' ') timfil='times.dat'
          inquire(file=timfil, exist=there)
          if (there) then
            ! Open the file with the times that each image was taken.
            call arkopn(iunit2, timfil, ' ', 'OLD', 'READONLY', &
            'FORMATTED', ' ', 0)
            ntim=0
            cttimes:  do
              read(iunit2, '(a)', iostat=iostat) timfil
              if (iostat < 0) exit cttimes
              read(iunit2, '(a)', iostat=iostat) timfil
              if (iostat < 0) exit cttimes
              ntim=ntim+1
            end do cttimes
            allocate(time(ntim))
            allocate(dtime(ntim))
            allocate(timnam(ntim))
            rewind (iunit2)
            do i=1, ntim
              read(iunit2, '(a)') timnam(i)
              read(iunit2, *) time(i), dtime(i)
              ! Remove leading blanks.
              deblank: do j=0, 50
                if (timnam(i)(1:1) .eq. ' ') then
                  timnam(i)=timnam(i)(2:50)
                else
                  exit deblank
                end if
              end do deblank
            end do
            close(iunit2)
            print*, '* ', ntim, ' times read.'
            exit times
          else
            print*, '* File not found.'
          end if
        end do times

        ! ***************************************************************
        !                   Read the list of bad pixels.
        ! ***************************************************************

        print*, ' '
        print*, '* The last file required is an X-Y file listing'
        print*, '* the bad pixels (again the first 3 lines are'
        print*, '* ignored).  Type "end" if you don''t have this file.'
        ! Prompt for the name of the bad pixel file, and return its size.
        i=sizark(nbpl)
        if (i .eq. -1) then  
          nbpl=0
        else
          allocate(ybpl(nbpl))
          allocate(xbpl(nbpl))
          ! Read the bad pixel file.
          i=inpark(nbpl, ybpl, xbpl)
        end if

        ! ***************************************************************
        !                   Set up all the options
        ! ***************************************************************

        read(iunit1, '(a)') ifname
        rewind (iunit1)
        ! Tell sizark what the name of the next file to be read is.
        call nxtark(ifname)
        ! Sizark returns the size of the file to be read.
        if (sizark(naxis) > 0) then
          inhigh=naxis
        else
          inhigh(1)=3000
          inhigh(2)=3000
        end if
        low(1)=1
        low(2)=1
        adu=6.7
        satur=16384.0
        call rd_detector(.true., .true., .true., .false., .true., &
        low, inhigh, adu, read, satur)

        if (clip_fwhm > 0.0) then
          clip: do 
            ! Most of the signal-to-noise is obtained by setting the clipping
            ! radius to be 2*fwhm.  But there is little gain in speed by
            ! making it less than 3 pixels.
            clip_fwhm=max(3.0/fwhm, 2.0)
            print*, &
            '> Give the radius for clipping the weight map in FWHM, <CR> = ', &
            clip_fwhm
            answer=' '
            read (*,'(a)') answer
            iostat=0
            if (answer .ne. ' ') read (answer,*,iostat=iostat) clip_fwhm
            if (iostat == 0) exit clip
          end do clip
        end if

        aper: do 
          ! Set the default aperture radius to the optimum (as shown in paper),
          ! but don't let it be less than 2 pixels.
          aprad=max(2.0, 2.0*fwhm/3.0)
          print*, &
          '> Give the radius for aperture photometry (<0 to skip), <CR> = ', &
          aprad
          answer=' '
          read (*,'(a)') answer
          iostat=0
          if (answer .ne. ' ') read (answer,*,iostat=iostat) aprad
          if (iostat == 0) exit aper
        end do aper

        if (clip_fwhm > 0.0) then
          optstar: do 
            iopt=nostar(1)
            print*, '* The extraction can be optimised for a particular star'
            print*, '* or for sky limited stars.'
            print*, '> Give the star number (or -1 for sky limit), <CR> = ', &
            iopt
            answer=' '
            read (*,'(a)') answer
            iostat=0
            if (answer .ne. ' ') read (answer,*,iostat=iostat) iopt
            if (iostat == 0) then
              if (iopt == -1) exit optstar
              do i=1, nstar
                if (nostar(i) == iopt) then
                  iopt=i
                  exit optstar
                end if
              end do
              print*, '* That star was not in the file, try again.'
            end if
          end do optstar
        end if

        ! Begin some output files.
        if (nstar < 100) then
          do i=1, nstar
            ofopt(i)='star00.opt'
 	    write(ofopt(i)(5:6),'(i2.2)') nostar(i)
            if (clip_fwhm > 0.0) then
              ! Create the output file for the optimal photometry of star nstar.
              call arkopn(iunit2, ofopt(i), ' ', 'NEW', 'OVERWRITE', &
              'FORMATTED', ' ', 0) 
              write(iunit2,*) '! Weight mapped clipped at ', clip_fwhm,  &
              ' times the FWHM.'
              write(iunit2,*) '! ', adu, ' detected photons per count.'
              if (dpos(i) > 0) then
                write(iunit2,*) '! Radius searched for star ', dpos(i), ' pixels.'
              else
                write(iunit2,*) '! Star position fixed.'
              end if
              close(unit=iunit2)
            end if
            if (aprad > 0.0) then
              ofap(i)=ofopt(i)
	      ofap(i)(8:10)='dat'
              ! Create the output file for the aperture photometry of star nstar.
              call arkopn(iunit4, ofap(i), ' ', 'NEW', 'OVERWRITE', &
              'FORMATTED', ' ', 0)
	      write(iunit4,*) '! Aperture of ', aprad, ' pixels radius.'
              write(iunit4,*) '! ', adu, ' detected photons per count.'
              if (dpos(i) > 0) then
                write(iunit4,*) '! Radius searched for star ', dpos(i), ' pixels.'
              else
                write(iunit4,*) '! Star position fixed.'
              end if
	      close(unit=iunit4)
            end if
          end do
        end if

        xpos_new=0.0
        ypos_new=0.0
        npos_new=0
        tot_sig_nos=0.0

        ! ***************************************************************
        !                   Now begin processing data.
        ! ***************************************************************

        frame: do

        !   *************************************************************
        !                    Go around this loop for each file.
        !   *************************************************************

          !  Reads in filename and offsets.
	  read(iunit1,'(a)',iostat=iostat) ifname
          if (iostat < 0) exit frame
	  read(iunit1,*,iostat=iostat) xoff, yoff, sky, theoff, ntoe, fwhm
          if (iostat < 0) exit frame
          theoff=theoff/57.2957795

          ! Reads in image.
	  print*, '********************************************'
          ! Tell sizark what the name of the next file to be read is.
          call nxtark(ifname)
          ! Get the size of the next file.
          i=sizark(naxis)
          if (i < 0) then
            print*, '* Going to next file.'
            cycle frame
          end if
          ! Allocate the arrays.
          allocate(data(naxis(1),naxis(2)))
          allocate(axdata( max(naxis(1),naxis(2)), 2))
          high=min(naxis,inhigh)
          ! Read in the data.  The image is contained in the array data.
          i=inpark(naxis, data, axdata)
          ! Recover the name of the last file read by lstark.
          call lstark(ifname)
          ! Remove leading blanks.
          frblank: do i=1, 50
            if (ifname(1:1) .eq. ' ') then
              ifname=ifname(2:50)
            else
              exit frblank
            end if
          end do frblank
                      
          ! Now make up an output log file name.  Strip off the extension.
          oflog=ifname
          extstr: do i=50, 2, -1
            if (oflog(i:i) .eq. '.') then
              oflog(i:50)=' '
              exit extstr
            end if
          end do extstr
          if (clip_fwhm > 0.0) then
            ! And open it for optimal photometry.
            call arkopn(logopt_unit, oflog, 'opt', 'NEW', 'OVERWRITE', &
            'FORMATTED', ' ', 0)
          end if
          if (aprad > 0.0) then
            ! And for the aperture photometry.
            call arkopn(logap_unit, oflog, 'dat', 'NEW', 'OVERWRITE', &
            'FORMATTED', ' ', 0)
          end if
                
          ! Try and match the file name to a known time.
          if (ntim /= 0) then
            do i=1, ntim
              if (ifname .eq. timnam(i)) then
                timec=time(i)
                dtimec=dtime(i)
                goto 560
              end if
            end do
            ! When this fails use the frame number.
            print*, 'Failed to find a time to match this frame.'
            print*, 'Will use the frame number.'
          end if       
          do i=1, 50
            if (ichar(ifname(i:i)).lt.48 .or. &
            ichar(ifname(i:i)).gt.57) ifname(i:i)=' '
          end do 
          timec=0.0
          read (ifname, *, iostat=iostat) timec
          dtimec=0.5
560       print*, 'Frame time is ', timec, '+/-', dtimec

          if (clip_fwhm > 0.0) then
            do i=1, npsf
              r1 = sqrt(xpsf(i)*xpsf(i) + ypsf(i)*ypsf(i))
              theta1=atan2(ypsf(i),xpsf(i))                                   
              xpos0(i)=r1*cos(theta1+theoff)+xoff
              ypos0(i)=r1*sin(theta1+theoff)+yoff
              if (debug) print*, '@ Estimated position of PSF star ', nopsf(i), &
              ' is ', xpos0(i), ypos0(i)
            end do

            ! Now fit the PSF star.
            call psf_calc(xpos0, ypos0, dpsf, npsf, data, &
            high, low, fwhm, clip_fwhm*fwhm, adu, satur, shape_par, ipsf)
            if (ipsf == -1) then
              print*, 'No good PSF star could be found within frame.'
              deallocate(data)                               
              deallocate(axdata)
              close (logopt_unit)
              if (aprad > 0.0) close (logap_unit)
              cycle frame
            end if
            print*, 'Fitted PSF star ', nopsf(ipsf), ' which gave FWsHM ', &
            1.665*shape_par(1), 1.665*shape_par(2)
            print*, 'Rotated at an angle of ', 57.29*shape_par(3), &
            ' degrees from the vertical.'
            cliprad = clip_fwhm*1.665*sqrt(shape_par(1)*shape_par(2))
            print*, 'Will use a clipping radius of ', cliprad, ' pixels.'
            write(logopt_unit,*) '! PSF star ', nopsf(ipsf), ' had FWsHM ', &
            1.665*shape_par(1), 1.665*shape_par(2)
            write(logopt_unit,*) '! Rotated at an angle of ', 57.29*shape_par(3), &
            ' degrees from the vertical.'
            write(logopt_unit,*) '! Clipping radius was ', cliprad, ' pixels.'
            ! Write to seeing.dat.
            write(iunit5,*) timec, sqrt(1.665*shape_par(1)*1.665*shape_par(2))
            if (aprad > 0.0) then
              write(logap_unit,*) '! Aperture of ', aprad, ' pixels radius.'
              write(logap_unit,*) '! ', adu, ' detected photons per count.'
              write(logap_unit,*) '!'
            end if
                           
            ! Now get a handle on the flux in the star to be extracted
            ! optimally.
            if (iopt > 0) then
              r1 = sqrt(xcoor(iopt)*xcoor(iopt) + ycoor(iopt)*ycoor(iopt))
              theta1=atan2(ycoor(iopt),xcoor(iopt))
              xpos=r1*cos(theta1+theoff)+xoff
              ypos=r1*sin(theta1+theoff)+yoff
              call extr(xpos, ypos, dpos(iopt), data, high, low, fwhm, cliprad, &
              adu, satur, shape_par, 0.0, .false., xcomp, ycomp, flux, error, &
              xfit, yfit, xerr, yerr, optnrm, best_n, iflag)
              if (iflag == -1) then
                print*, 'Star to be optimised is too close to frame edge.'
                print*, 'Going to next frame.'
                deallocate(data)                               
                deallocate(axdata)
                close (logopt_unit)
                if (aprad > 0.0) close (logap_unit)
                cycle frame
              end if
              if (iflag == -4) then
                print*, 'Star to be optimised is saturated.'
                print*, 'Going to next frame.'
                deallocate(data)                               
                deallocate(axdata)
                close (logopt_unit)
                if (aprad > 0.0) close (logap_unit)
                cycle frame
              end if
              if (iflag == -3) then
                print*, 'Failed to fit sky for star to be optimised.'
                print*, 'Going to next frame.'
                deallocate(data)                               
                deallocate(axdata)        
                close (logopt_unit)
                if (aprad > 0.0) close (logap_unit)
                cycle frame
              end if
              print*, 'Extractions will be optimised for stars with ', &
              optnrm, ' peak counts.'
            else                            
              optnrm=0.0
            end if

          else
            
            ! Don't do optimal photometry, but we will need a FWHM for fitting
            ! the star to find its position.
            shape_par(1)=fwhm/1.665
            shape_par(2)=fwhm/1.665
            shape_par(3)=0.0
            ! And a cliprad.
            cliprad=2.0*fwhm
            ! And (believe it or not) a star flux to be optimised.
            optnrm = 0.0
          end if
            

          star: do istar=1, nstar                   
            ! ***********************************************************
            !                Go around this loop for each star.
            ! ***********************************************************

            ! Calculate approximate position of star.
            r1 = sqrt(xcoor(istar)*xcoor(istar) + ycoor(istar)*ycoor(istar))
            theta1=atan2(ycoor(istar),xcoor(istar))
            xpos=r1*cos(theta1+theoff)+xoff
            ypos=r1*sin(theta1+theoff)+yoff

            ! Search for nearby stars, in the co-ordinates of the reference
            ! frame.
            comp=.false.
            companion: do jstar=1, nstar
              if (istar == jstar) cycle companion
              if ( (xcoor(istar)-xcoor(jstar))**2.0 + &
                   (ycoor(istar)-ycoor(jstar))**2.0 < (3.0*fwhm)**2.0 ) then
                print*, 'Target star ', nostar(istar), &
                ' will be fitted with companion ', nostar(jstar)
                ! Set the position.
                r1 = sqrt(xcoor(jstar)*xcoor(jstar) + ycoor(jstar)*ycoor(jstar))
                theta1=atan2(ycoor(jstar),xcoor(jstar))
                xcomp=r1*cos(theta1+theoff)+xoff
                ycomp=r1*sin(theta1+theoff)+yoff
                comp=.true.
                exit companion
              end if
            end do companion

            call extr(xpos, ypos, dpos(istar), data, high, low, fwhm, cliprad, &
            adu, satur, shape_par, optnrm, comp, xcomp, ycomp, flux, error, &
            xfit, yfit, xerr, yerr, peak, best_n, iflag)

            ! Check for bad pixels, first set nearest bad pixel to
            ! be closest edge of the frame.
            badpix=min(xfit, real(naxis(1))-xfit, &
            yfit, real(naxis(2))-yfit )
            badpix = badpix + 1.0
            if (nbpl /= 0) then
              do i=1, nbpl
                badpix = min (sqrt( &
                (xfit-xbpl(i))*(xfit-xbpl(i)) + &
                (yfit-ybpl(i))*(yfit-ybpl(i)) ), badpix)
              end do
            end if
            if (iflag == 0) then
              if (badpix < cliprad+0.5) iflag=max(nint(badpix),1)
            else if (iflag == -1) then
              iflag = -2
            end if

            if (clip_fwhm > 0.0) then
              if (iflag == 0) then
                print*, 'Target star ', nostar(istar), ' gave ', flux, &
                ' counts +/- ', 100.0*error/abs(flux), '%'
              else if (iflag == -3) then
                print*, 'Failed to fit sky for target star ', nostar(istar)
              else if (iflag == -4) then
                print*, 'Star ', nostar(istar), ' is saturated.'
              else if (iflag == -5) then
                print*, 'Star ', nostar(istar), ' has ill-determined sky.'
              else if (iflag > 0) then
                print*, 'Star ', nostar(istar), ' has a bad pixel.'
              end if
              if (nstar < 100) then
                ! Write it to the appropriate star file.
                call arkopn(iunit2, ofopt(istar), ' ', 'OLD', &
                'APPEND', 'FORMATTED', ' ', 0)
                write(iunit2,*) timec, flux, error, dtimec, iflag
                close(iunit2)
              end if
              ! Write to the frame's log file.
              write(logopt_unit,*) nostar(istar), xfit, xerr, yfit, yerr
              write(logopt_unit,*) flux, error, iflag
            end if

            if (iflag == -2) then
              print*, 'Star ', nostar(istar), ' is too close to frame edge.'
              goto 850
            end if

            if (dpos(istar) > 0.0) then
              print*, 'Fitted position of star ', nostar(istar), &
              sqrt((xpos-xfit)**2.0 + (ypos-yfit)**2.0), &
              'from start position.'
            end if

            if (iflag == 0) then
              ! Keep a track on the mean position of the star in all frames.
              xpos_new(istar)=xpos_new(istar)+xfit-xoff
              ypos_new(istar)=ypos_new(istar)+yfit-yoff
              npos_new(istar)=npos_new(istar)+1
              if (debug) print*, '@ Star is at ', xfit, yfit
              ! And on the signal-to-noise.
              tot_sig_nos(istar)=tot_sig_nos(istar) + 100.0*best_n/error
            end if

            ! And the aperture photometry.  
            if (aprad > 0.0) then
              sky_stuffed = .false.
              ! The box size is determined in the way described in the paper.
              ibox=int(sqrt(628.4*aprad*aprad+ 4.0*fwhm*fwhm))
              call skyfit(data, xfit, yfit, fwhm, ibox, low, high, &
              skycnt, skyerr, skynos, iflag)
              if (iflag == -5) sky_stuffed = .true.
              if (iflag == -3) then
                print*, 'Failed to fit sky for aperture photometry for target star ', &
                nostar(istar)
                flux = 0.0
                error = 0.0
              else
                flux=sum_aper(data, aprad, skycnt, skyerr, skynos, &
                xfit, yfit, adu, satur, low, high, error, iflag)
                if (iflag == 0) then
                  if (badpix < cliprad+0.5) iflag=max(nint(badpix), 1)
                end if              
                if (clip_fwhm < 0.0) then
                  if (iflag == 0) then
                    print*, 'Target star ', nostar(istar), ' gave ', flux, &
                    ' counts +/- ', 100.0*error/abs(flux), '%'
                  else if (iflag == -1) then
                    print*, 'Target star ', nostar(istar), &
                    ' is too close to frame edge for aperture photometry.'
                  else if (iflag == -4) then
                    print*, 'Warning, target star ', nostar(istar), &
                    ' is saturated for aperture photometry.'
                  end if
                end if
                if (sky_stuffed) iflag = -5
              end if
            end if

850         if (aprad > 0.0) then
              if (nstar < 100) then
                call arkopn(iunit4, ofap(istar), ' ', 'OLD', &
                'APPEND', 'FORMATTED', ' ', 0)
                write(iunit4,*) timec, flux, error, dtimec, iflag
                close(iunit4)
              end if
              write(logap_unit,*) nostar(istar), xfit, xerr, yfit, yerr
              write(logap_unit,*) flux, error, iflag
            end if

          end do star

          deallocate(data)                               
          deallocate(axdata)
          if (clip_fwhm > 0.0) close (logopt_unit)
          if (aprad > 0.0) close (logap_unit)

        end do frame

        close(iunit1)
        if (clip_fwhm > 0.0) close(iunit5)

        ! Finally, write a file of updated star positions.
        call arkopn(iunit1, 'stars.pos_new', ' ', 'NEW', 'OVERWRITE', &
        'FORMATTED', ' ', 0)
        write(iunit1,*)
        write(iunit1,*)
        write(iunit1,*)
        do istar=1, nstar              
          if (npos_new(istar) > 0) then
            write(iunit1,*) nostar(istar), &
            xpos_new(istar)/real(npos_new(istar)), &
            ypos_new(istar)/real(npos_new(istar)), 0.0
            print*, '% of available S:N extracted for star ', &
            nostar(istar), '=', tot_sig_nos(istar)/real(npos_new(istar))
          end if
        end do        

      end program opphot
